"""Multi-stage optimization workflow."""

import re
from typing import Self, Sequence

import more_itertools as mit
from pydantic import BaseModel, Field, model_validator

from stjames.correction import Correction

from ..constraint import Constraint
from ..method import XTB_METHODS, Method
from ..mode import Mode
from ..opt_settings import OptimizationSettings
from ..settings import Settings
from ..solvent import Solvent, SolventModel, SolventSettings
from ..task import Task
from ..types import UUID
from .workflow import MoleculeWorkflow


class MultiStageOptSettings(BaseModel):
    """
    Settings for multi-stage optimizations.

    RECKLESS
        GFN2-xTB//GFN-FF (no pre-opt)
    RAPID *default
        r²SCAN-3c//GFN2-xTB with GFN0-xTB pre-opt (off by default)
    CAREFUL
        wB97X-3c//r²SCAN-3c with GFN2-xTB pre-opt
    METICULOUS
        wB97M-D3BJ/def2-TZVPPD//wB97X-3c//r²SCAN-3c with GFN2-xTB pre-opt

    Notes:
    - No solvent in any optimizations when using Modes
    - If solvent: xTB singlepoints use CPCMX, xTB optimizations use ALBP, all else use CPCM
    - Allows a single point to be called with no optimization

    :param mode: Mode for settings
    :param optimization_settings: list of opt settings to apply successively
    :param singlepoint_settings: final single point settings
    :param solvent: solvent to use for singlepoint
    :param xtb_preopt: pre-optimize with xtb (sets based on mode when None)
    :param constraints: constraints for optimization
    :param transition_state: whether this is a transition state
    :param frequencies: whether to calculate frequencies

    >>> msos = MultiStageOptSettings(mode=Mode.RAPID, solvent="water")
    >>> msos
    <MultiStageOptSettings RAPID>
    >>> msos.level_of_theory
    'r2scan_3c/cpcm(water)//gfn2_xtb'
    """

    mode: Mode
    optimization_settings: Sequence[Settings] = tuple()
    singlepoint_settings: Settings | None = None
    solvent: Solvent | None = None
    xtb_preopt: bool = False
    constraints: Sequence[Constraint] = tuple()
    transition_state: bool = False
    frequencies: bool = False

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        """
        String representation of the settings.

        >>> print(MultiStageOptSettings(mode=Mode.RAPID, solvent="water"))
        <MultiStageOptSettings RAPID>
        """
        if self.mode != Mode.MANUAL:
            return f"<{type(self).__name__} {self.mode.name}>"

        return f"<{type(self).__name__} {self.level_of_theory}>"

    @property
    def level_of_theory(self) -> str:
        """
        Returns the level of theory for the workflow.

        >>> msos = MultiStageOptSettings(mode=Mode.RAPID, solvent="hexane")
        >>> msos.level_of_theory
        'r2scan_3c/cpcm(hexane)//gfn2_xtb'
        """
        methods = [self.singlepoint_settings] if self.singlepoint_settings else []
        methods += reversed(self.optimization_settings)

        return "//".join(m.level_of_theory for m in methods)

    @model_validator(mode="after")
    def set_mode_and_settings(self) -> Self:
        """Check mode and settings are properly specified, and assign settings based on mode."""
        opt_or_sp = bool(self.optimization_settings) or bool(self.singlepoint_settings)

        if self.mode == Mode.AUTO:
            self.mode = Mode.RAPID

        match (self.mode, opt_or_sp):
            case (Mode.DEBUG, _):
                raise NotImplementedError("Unsupported mode: DEBUG")

            case (Mode.MANUAL, False):
                raise ValueError("Must specify at least one of optimization_settings or singlepoint_settings with MANUAL mode")
            case (Mode.MANUAL, True):
                pass

            case (mode, True):
                raise ValueError(f"Cannot specify optimization_settings or singlepoint_settings with {mode=}")

            case (mode, False):
                self._assign_settings_by_mode(mode)

        return self

    def _assign_settings_by_mode(self, mode: Mode) -> None:
        """
        Construct the settings based on the Mode.

        :param mode: Mode to use
        """
        opt_settings = OptimizationSettings(constraints=self.constraints, transition_state=self.transition_state)

        # No solvent in pre-opt
        OPT = [Task.OPTIMIZE if not self.transition_state else Task.OPTIMIZE_TS]
        gfn0_pre_opt = [Settings(method=Method.GFN0_XTB, tasks=OPT, mode=Mode.RAPID, opt_settings=opt_settings)]
        gfn2_pre_opt = [Settings(method=Method.GFN2_XTB, tasks=OPT, mode=Mode.RAPID, opt_settings=opt_settings)]

        def opt(method: Method, basis_set: str | None = None, solvent: Solvent | None = None, freq: bool = False) -> Settings:
            """Generates optimization settings."""
            model = "alpb" if method in XTB_METHODS else "cpcm"

            return Settings(
                method=method,
                basis_set=basis_set,
                tasks=OPT + [Task.FREQUENCIES] * freq,
                solvent_settings=SolventSettings(solvent=solvent, model=model) if solvent else None,
                opt_settings=opt_settings,
            )

        def sp(method: Method, basis_set: str | None = None, solvent: Solvent | None = None) -> Settings:
            """Generate singlepoint settings."""
            model = "cpcmx" if method in XTB_METHODS else "cpcm"

            return Settings(
                method=method,
                basis_set=basis_set,
                solvent_settings=SolventSettings(solvent=solvent, model=model) if solvent else None,
            )

        match mode:
            case Mode.RECKLESS:
                self.xtb_preopt = False
                self.optimization_settings = [opt(Method.GFN_FF, freq=self.frequencies)]
                self.singlepoint_settings = sp(Method.GFN2_XTB, solvent=self.solvent)

            case Mode.RAPID:
                self.optimization_settings = [
                    *gfn0_pre_opt * self.xtb_preopt,
                    opt(Method.GFN2_XTB, freq=self.frequencies),
                ]
                self.singlepoint_settings = sp(Method.R2SCAN3C, solvent=self.solvent)

            case Mode.CAREFUL:
                self.optimization_settings = [
                    *gfn2_pre_opt * self.xtb_preopt,
                    opt(Method.R2SCAN3C, freq=self.frequencies),
                ]
                self.singlepoint_settings = sp(Method.WB97X3C, solvent=self.solvent)

            case Mode.METICULOUS:
                self.optimization_settings = [
                    *gfn2_pre_opt * self.xtb_preopt,
                    opt(Method.R2SCAN3C),
                    opt(Method.WB97X3C, freq=self.frequencies),
                ]
                self.singlepoint_settings = sp(Method.WB97MD3BJ, "def2-TZVPPD", solvent=self.solvent)

            case mode:
                raise NotImplementedError(f"Cannot assign settings for {mode=}")


class MultiStageOptWorkflow(MoleculeWorkflow, MultiStageOptSettings):
    """
    MoleculeWorkflow for multi-stage optimizations.

    Inherited
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow
    :param optimization_settings: list of opt settings to apply successively
    :param singlepoint_settings: final single point settings
    :param solvent: solvent to use for singlepoint
    :param xtb_preopt: pre-optimize with xtb (sets based on mode when None)
    :param constraints: constraints for optimization
    :param transition_state: whether this is a transition state
    :param frequencies: whether to calculate frequencies

    Populated while running
    :param calculations: list of calculation UUIDs

    >>> from stjames.molecule import Atom, Molecule
    >>> He = Molecule(charge=0, multiplicity=1, atoms=[Atom(atomic_number=2, position=[0, 0, 0])])
    >>> msow = MultiStageOptWorkflow(initial_molecule=He, mode=Mode.RAPID, solvent="water")
    >>> msow
    <MultiStageOptWorkflow RAPID>
    >>> msow.level_of_theory
    'r2scan_3c/cpcm(water)//gfn2_xtb'
    """

    # Populated while running the workflow
    calculations: list[UUID | None] = Field(default_factory=list)

    def __repr__(self) -> str:
        if self.mode != Mode.MANUAL:
            return f"<{type(self).__name__} {self.mode.name}>"

        return f"<{type(self).__name__} {self.level_of_theory}>"


# the id of a mutable object may change, thus using object()
_sentinel_msos = object()


class MultiStageOptMixin(BaseModel):
    """
    Mixin for workflows that use MultiStageOptSettings.
    """

    mso_mode: Mode = Mode.AUTO
    # Need to use a sentinel object to make both mypy and pydantic happy
    multistage_opt_settings: MultiStageOptSettings = _sentinel_msos  # type: ignore [assignment]
    solvent: Solvent | None = None
    xtb_preopt: bool = False
    constraints: Sequence[Constraint] = tuple()
    transition_state: bool = False
    frequencies: bool = False

    @model_validator(mode="after")
    def set_mso_settings(self) -> Self:
        if self.mso_mode == Mode.AUTO:
            self.mso_mode = Mode.RAPID

        match self.mso_mode, self.multistage_opt_settings:
            case (Mode.DEBUG, _):
                raise NotImplementedError("Unsupported mode: DEBUG")

            case (Mode.MANUAL, msos) if msos is _sentinel_msos:
                raise ValueError("Must specify multistage_opt_settings with MANUAL mode")
            case (Mode.MANUAL, _):
                pass

            case (mso_mode, msos) if msos is not _sentinel_msos:
                raise ValueError(f"Cannot specify multistage_opt_settings with {mso_mode=}, {msos=}")

            case (mso_mode, _):
                self.multistage_opt_settings = MultiStageOptSettings(
                    mode=mso_mode,
                    solvent=self.solvent,
                    xtb_preopt=self.xtb_preopt,
                    constraints=self.constraints,
                    transition_state=self.transition_state,
                    frequencies=self.frequencies,
                )

        return self


def mso_settings_from_method_string(
    methods: str,
    solvent: Solvent | None = None,
    constraints: list[Constraint] | None = None,
    transition_state: bool = False,
    frequencies: bool = False,
) -> MultiStageOptSettings:
    """
    Helper function to construct multi-stage opt settings objects from a method string.

    >>> mso_settings_from_method_string("r2SCAN-3c/CPCM(Water)//B3LYP-D3/6-31G(d)/ALPB(Water)//GFN2-xTB/CPCM(Water)//GFN0-xTB").level_of_theory
    'r2scan_3c/cpcm(water)//b3lyp-d3/6-31g(d)/alpb(water)//gfn2_xtb/cpcm(water)//gfn0_xtb'
    """
    solvent_models = "|".join(model.name for model in SolventModel)

    pattern = rf"""
        (?P<method>[^/()]+)                                 # Method + optional corrections
        (?:/(?P<basis_set>(?!{solvent_models})[^/]+?))?     # Optional basis_set, not starting with solvent model name
        (?:/(?P<solvent_model>{solvent_models})             # Optional solvent model
            \((?P<solvent>[^()]+)\))?                       # Solvent name in parentheses
        (?:\/\/|$)                                          # End or separator
"""
    constraints = constraints or []
    opt_settings = OptimizationSettings(constraints=constraints, transition_state=transition_state)
    OPT = [Task.OPTIMIZE if not transition_state else Task.OPTIMIZE_TS]

    valid_corrections = {c.name.lower() for c in Correction}  # Python3.11 hack

    def process(match: re.Match[str]) -> Settings:
        data = match.groupdict()

        method, corrections = mit.partition(lambda x: x.lower() in valid_corrections, data["method"].split("-"))
        solvent_settings = SolventSettings(solvent=data["solvent"], model=data["solvent_model"]) if data["solvent"] else None

        return Settings(
            method="-".join(method),
            basis_set=data["basis_set"],
            tasks=OPT,
            solvent_settings=solvent_settings,
            opt_settings=opt_settings,
            corrections=list(corrections),
        )

    optimization_settings = [process(match) for match in re.finditer(pattern, methods, re.VERBOSE | re.IGNORECASE)]
    if len(optimization_settings) > 1:
        sp_settings = optimization_settings.pop(0)
        sp_settings.tasks = [Task.ENERGY]
    else:
        sp_settings = None

    optimization_settings = optimization_settings[::-1]
    if frequencies:
        optimization_settings[-1].tasks.append(Task.FREQUENCIES)

    return MultiStageOptSettings(
        mode=Mode.MANUAL,
        optimization_settings=optimization_settings,
        singlepoint_settings=sp_settings,
        solvent=solvent,
        xtb_preopt=False,
        constraints=constraints,
        transition_state=transition_state,
        frequencies=frequencies,
    )


def build_mso_settings(
    sp_method: Method,
    sp_basis_set: str | None,
    opt_methods: list[Method],
    opt_basis_sets: list[str | None],
    mode: Mode = Mode.MANUAL,
    solvent: Solvent | None = None,
    use_solvent_for_opt: bool = False,
    constraints: list[Constraint] | None = None,
    transition_state: bool = False,
    frequencies: bool = False,
) -> MultiStageOptSettings:
    """
    Helper function to construct multi-stage opt settings objects manually.

    There's no xTB pre-optimization here - add that yourself!

    :param optimization_settings: optimization settings to apply successively
    :param singlepoint_settings: final single point settings
    :param mode: Mode for settings, defaults to `MANUAL`
    :param solvent: solvent to use
    :param use_solvent_for_opt: whether to conduct opts with solvent
    :param constraints: constraints for optimization
    :param transition_state: whether this is a transition state
    :param frequencies: whether to calculate frequencies
    :returns: MultiStageOptSettings
    """
    if constraints is None:
        constraints = []

    opt_settings = OptimizationSettings(constraints=constraints, transition_state=transition_state)

    OPT = [Task.OPTIMIZE if not transition_state else Task.OPTIMIZE_TS]

    def opt(method: Method, basis_set: str | None = None, solvent: Solvent | None = None, freq: bool = False) -> Settings:
        """Generates optimization settings."""
        model = "alpb" if method in XTB_METHODS else "cpcm"

        return Settings(
            method=method,
            basis_set=basis_set,
            tasks=OPT + [Task.FREQUENCIES] * freq,
            solvent_settings=SolventSettings(solvent=solvent, model=model) if (solvent and use_solvent_for_opt) else None,
            opt_settings=opt_settings,
        )

    def sp(method: Method, basis_set: str | None = None, solvent: Solvent | None = None) -> Settings:
        """Generate singlepoint settings."""
        model = "cpcmx" if method in XTB_METHODS else "cpcm"

        return Settings(
            method=method,
            basis_set=basis_set,
            tasks=[Task.ENERGY],
            solvent_settings=SolventSettings(solvent=solvent, model=model) if solvent else None,
        )

    return MultiStageOptSettings(
        mode=mode,
        optimization_settings=[
            opt(method=method, basis_set=basis_set, solvent=solvent, freq=frequencies) for method, basis_set in zip(opt_methods, opt_basis_sets, strict=True)
        ],
        singlepoint_settings=sp(method=sp_method, basis_set=sp_basis_set, solvent=solvent),
        solvent=solvent,
        xtb_preopt=False,
        constraints=constraints,
        transition_state=transition_state,
        frequencies=frequencies,
    )


def multi_stage_opt_settings_from_workflow(msow: MultiStageOptWorkflow) -> MultiStageOptSettings:
    """
    Helper function to convert a MultiStageOptWorkflow to MultiStageOptSettings.

    :param msow: MultiStageOptWorkflow
    :returns: MultiStageOptSettings

    >>> from stjames.molecule import Atom, Molecule
    >>> He = Molecule(charge=0, multiplicity=1, atoms=[Atom(atomic_number=2, position=[0, 0, 0])])
    >>> msos = multi_stage_opt_settings_from_workflow(
    ...     MultiStageOptWorkflow(initial_molecule=He, mode=Mode.RAPID, solvent="water")
    ... )
    >>> print(msos)
    <MultiStageOptSettings RAPID>
    >>> msos.level_of_theory
    'r2scan_3c/cpcm(water)//gfn2_xtb'
    >>> msos.solvent
    <Solvent.WATER: 'water'>
    >>> msos.xtb_preopt
    False
    """
    data = dict(msow)
    del data["calculations"]
    return MultiStageOptSettings.construct(**data)



from typing import Annotated, Self

from pydantic import AfterValidator, PositiveFloat, PositiveInt, model_validator

from ..base import Base, round_float
from ..pdb import PDB
from ..types import UUID, round_list
from .workflow import MoleculeWorkflow


class BindingPoseContact(Base):
    """
    A single protein–ligand contact from an MD trajectory.

    :param protein_atom_index: the index of the protein atom
    :param ligand_atom_index: the index of the ligand atom
    :occupancy: the probability of seeing this interaction in a frame, between 0 and 1
    """

    protein_atom_index: int
    ligand_atom_index: int
    occupancy: Annotated[float, AfterValidator(round_float(3))]


class BindingPoseTrajectory(Base):
    """
    Represents a single trajectory looking at a binding pose.

    :param uuid: the UUID of the trajectory
    :param ligand_rmsd: the RMSD of the ligand vs. starting pose (aligning the protein)
    :param contacts: the conserved binding-pose contacts
    """

    uuid: UUID
    ligand_rmsd: Annotated[list[float], AfterValidator(round_list(3))] = []
    contacts: list[BindingPoseContact] = []


class PoseAnalysisMolecularDynamicsWorkflow(MoleculeWorkflow):
    """
    Pose analysis molecular dynamics workflow.

    Note that the protein can be supplied either by UUID or raw PDB object.
    We anticipate that the former will dominate deployed usage, but the latter is handy for isolated testing.
    If, for whatever reason, the workflow is initialized with both a `target_uuid` and a `target`, the UUID will be ignored.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow (currently unused)

    New:
    :param protein: PDB of the protein.
    :param protein_uuid: UUID of the protein.
    :param num_trajectories: how many trajectories to run
    :param equilibration_time_ns: how long to equilibrate trajectories for, in nanoseconds
    :param simulation_time_ns: how long to run trajectories for, in nanoseconds
    :param temperature: the temperature, in K
    :param pressure_atm: the pressure, in atm
    :param langevin_timescale_ps: the timescale for the Langevin integrator, in inverse picoseconds
    :param timestep_fs: the timestep, in femtoseconds
    :param constrain_hydrogens: whether or not to use SHAKE to freeze bonds to hydrogen
    :param nonbonded_cutoff: the nonbonded cutoff for particle-mesh Ewald, in Å
    :param protein_prune_cutoff: the cutoff past which residues will be deleted, in Å
    :param protein_restraint_cutoff: the cutoff past which alpha-carbons will be constrained, in Å
    :param protein_restraint_constant: the force constant for backbone restraints, in kcal/mol/Å**2
    :param ionic_strength_M: the ionic strength of the solution, in M (molar)
    :param water_buffer: the amount of water to add around the protein, in Å

    Results:
    :param trajectories: for each replicate, a UUID and the corresponding analysis results
    :param minimized_protein_uuid: UUID of final system PDB
    """

    protein: PDB | None = None
    protein_uuid: UUID | None = None

    num_trajectories: PositiveInt = 4
    equilibration_time_ns: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 5
    simulation_time_ns: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 10

    temperature: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 300
    pressure_atm: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 1.0
    langevin_timescale_ps: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 1.0

    timestep_fs: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 2
    constrain_hydrogens: bool = True
    nonbonded_cutoff: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 8.0

    protein_prune_cutoff: Annotated[PositiveFloat, AfterValidator(round_float(3))] | None = None
    protein_restraint_cutoff: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 7.0
    protein_restraint_constant: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 100

    ionic_strength_M: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 0.10
    water_buffer: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 6.0

    minimized_protein_uuid: UUID | None = None
    bonds: list[tuple[int, int]] = []
    trajectories: list[BindingPoseTrajectory] = []

    @model_validator(mode="after")
    def check_cutoff_sanity(self) -> Self:
        """Check if protein is provided."""
        if self.protein_prune_cutoff is not None:
            if self.protein_prune_cutoff < self.protein_restraint_cutoff:
                raise ValueError("Pruning cutoff must be larger than restraint cutoff")
        return self



"""Ion mobility workflow."""

from ..types import UUID
from .workflow import MoleculeWorkflow


class IonMobilityWorkflow(MoleculeWorkflow):
    """
    Workflow for calculating hydrogen bond basicity.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow (currently unused)

    New:
    :param protonate: automatically protonate the molecule
    :param temperature: the temperature, in Kelvin
    :param do_csearch: whether to perform a conformational search
    :param do_optimization: whether to perform an optimization

    Results:
    :param conformers: the UUIDs of the conformers
    :param conformer_ccs: the collision cross section (Å**2) per conformer
    :param conformer_ccs_stdev: the uncertainty in the same
    :param conformer_weights: the Boltzmann weights at RT
    :param average_ccs: the Boltzmann-weighted CCS for the ensemble
    :param average_ccs_stdev: the uncertainty in the same
    """

    protonate: bool = False
    temperature: float = 300
    do_csearch: bool = True
    do_optimization: bool = True

    conformers: list[UUID] = []

    conformer_ccs: list[float] = []
    conformer_ccs_stdev: list[float] = []
    boltzmann_weights: list[float] = []

    average_ccs: float | None = None
    average_ccs_stdev: float | None = None



"""Intrinsic reaction coordinate (IRC) workflow."""

from typing import Self

from pydantic import Field, PositiveFloat, field_validator, model_validator

from ..method import XTB_METHODS, Method
from ..mode import Mode
from ..settings import Settings
from ..solvent import Solvent, SolventModel, SolventSettings
from ..types import UUID
from .workflow import MoleculeWorkflow

_sentinel_settings: Settings = object()  # type: ignore [assignment]


class IRCWorkflow(MoleculeWorkflow):
    """
    Workflow for Intrinsic Reaction Coordinate (IRC) calculations.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow

    New:
    :param settings: Settings for running the IRC (only for manual mode)
    :param solvent: Solvent for the calculation (non-Manual mode only)
    :param preopt: whether to optimize the geometry before starting the IRC
    :param max_irc_steps: maximum number of steps for the IRC
    :param step_size: step size for the IRC (Å)

    Results:
    :param starting_TS: optimized TS before the IRC (==initial_molecule if preopt=False)
    :param irc_forward: forward calculations
    :param irc_backward: reverse calculations
    """

    settings: Settings = _sentinel_settings
    solvent: Solvent | None = None

    preopt: bool = False
    max_irc_steps: int = 10
    step_size: PositiveFloat = 0.05

    starting_TS: UUID | None = None

    irc_forward: list[UUID] = Field(default_factory=list)
    irc_backward: list[UUID] = Field(default_factory=list)

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        """String representation of the workflow."""
        if self.mode != Mode.MANUAL:
            return f"<{type(self).__name__} {self.mode.name}>"

        return f"<{type(self).__name__} {self.level_of_theory}>"

    @property
    def level_of_theory(self) -> str:
        """Level of theory for the workflow."""
        return self.settings.level_of_theory

    @field_validator("step_size", mode="after")
    @classmethod
    def validate_step_size(cls, step_size: float) -> float:
        """Validate the step size."""
        if step_size < 1e-3 or step_size > 0.1:
            raise ValueError(f"Step size must be between 0.001 and 0.1 Å, got: {step_size}")

        return step_size

    @model_validator(mode="after")
    def validate_mode(self) -> Self:
        """Convert the mode to settings."""
        if self.mode == Mode.MANUAL:
            if self.settings is _sentinel_settings:
                raise ValueError("Settings are required for manual calculations.")
            return self
        elif self.settings is not _sentinel_settings:
            raise ValueError("Cannot specify settings and mode simultaneously.")

        match self.mode:
            case Mode.RAPID:
                method = Method.GFN2_XTB
            case Mode.CAREFUL:
                method = Method.R2SCAN3C
            case Mode.METICULOUS:
                method = Method.WB97X3C
            case _:
                raise ValueError(f"Unsupported mode: {self.mode}")

        model = SolventModel.ALPB if method in XTB_METHODS else SolventModel.CPCM
        solvent_settings = SolventSettings(solvent=self.solvent, model=model) if self.solvent else None

        self.settings = Settings(mode=Mode.RAPID, method=method, solvent_settings=solvent_settings)

        return self



"""Protein Cofolding Workflow."""

from typing import Annotated

from pydantic import AfterValidator, BaseModel, ConfigDict

from ..base import LowercaseStrEnum, round_float
from ..types import UUID, round_optional_list
from .workflow import FASTAWorkflow


class CofoldingModel(LowercaseStrEnum):
    """Cofolding model to be used for prediction."""

    CHAI_1R = "chai_1r"
    BOLTZ_1 = "boltz_1"
    BOLTZ_2 = "boltz_2"


class CofoldingScores(BaseModel):
    confidence_score: Annotated[float, AfterValidator(round_float(3))]
    ptm: Annotated[float, AfterValidator(round_float(3))]  # predicted template modeling score
    iptm: Annotated[float, AfterValidator(round_float(3))]  # interface predicted template modeling score
    avg_lddt: Annotated[float, AfterValidator(round_float(3))]


class AffinityScore(BaseModel):
    pred_value: Annotated[float, AfterValidator(round_float(3))]
    probability_binary: Annotated[float, AfterValidator(round_float(3))]
    pred_value1: Annotated[float, AfterValidator(round_float(3))]
    probability_binary1: Annotated[float, AfterValidator(round_float(3))]
    pred_value2: Annotated[float, AfterValidator(round_float(3))]
    probability_binary2: Annotated[float, AfterValidator(round_float(3))]


class ProteinCofoldingWorkflow(FASTAWorkflow):
    """
    A workflow for predicting structures. Especially protein structures.

    Inherited:
    :param initial_protein_sequences: protein sequences of interest
    :param initial_smiles_list: SMILES strings of interest

    New:
    :param use_msa_server: whether to use the MSA server
    :param use_templates_server: whether to use the templates server
    :param predicted_structure_uuid: UUID of the predicted structure
    """

    model_config = ConfigDict(validate_assignment=True)

    use_msa_server: bool = False
    use_templates_server: bool = False
    use_potentials: bool = False
    predicted_structure_uuid: UUID | None = None
    scores: CofoldingScores | None = None
    model: CofoldingModel = CofoldingModel.BOLTZ_2
    affinity_score: AffinityScore | None = None
    lddt: Annotated[list[float] | None, AfterValidator(round_optional_list(3))] = None
